# -*- coding: utf-8 -*-
"""
Created on Wed Sep  9 21:30:50 2020

@author: oleh.melnyk
"""

import numpy as np

from scipy.sparse.linalg import eigsh


def stitch(objects, wtype='weighted', threshold=1e-12):
    '''
    In paralelized ptychography, each node returns recovered
    part of the object up to a phase factor. Our goal is to stich those pieces by synchronizing
    phase factors. To do so, we solve the angular synchronization, a
    recovery of the unknown phase factors from their (noisy) pairwise differences
    To obtain these pairwise differences, we compute the inner product between
    two overlapping parts. Since on the overlap two objects are approximately the
    same, it gives us the phase difference. Mathematically, it is given by
    < e^(i a_r1) O_r1, e^(i a_r2) O_r2 > approx. = e^(i [a_r1 - a_r2]) norm(O_r1,r_2)^2,
    where O_r is recovered r-th part of the object,
    O_r1,r2 - part of the object corresponging to the overlap of two parts,
    a_r - unknown phase factor.

    This inner products are stored in the ph_diff matrix. Then, the matrix is entrywise
    normalized to separate phases e^(i [a_r1 - a_r2]) and non-negative norm(O_r1,r_2)^2.

    In angular synchronization we take phase differences as data and
    can use norms as weights to solve (un)weighted least squares problem.
    Problem itself is NP-hard and, thus, its relaxation instead is solved, which is
    simply a computation of the eigenvector corresponding to the smallest eigenvalue
    of the graph laplacian matrix.

    Obtained eigenvector (v in code) is entrywise normalized to get phases.
    Each part of the object is then phase-rotated correspondingly. Finally, we
    join the pieces, avaraging on the overlaps.

    Note: obtained solution is the original object up to a globabl phase factor.

    For more details see Section 2 of https://arxiv.org/pdf/2005.02032.pdf

    Possible speed up of the routine: avoid dot product calculation of between all
    parts, but only those which overlap.

    Parameters
    ----------
    objects : numpy.ndarray
        d1 X d2 X R array containing recovered pieces of object,
        required to be stitched
    wtype : str, optional
        Selected type of weights to be used in the angular synchronization
        There are 2 possible choices:
        'unweighted' or 'weighted', depending on usage or not of the weights for
        angular synchronization (explained above)
    threshold : float, optional
        Minimum absolute value to be considered in finding the phase match for stitching.
'''
    d1 = objects.shape[0]
    d2 = objects.shape[1]
    R = objects.shape[2]

    ph_diff = np.zeros((R, R), dtype=complex)

    for r1 in range(R):
        for r2 in range(r1+1, R):
            ph_diff[r1, r2] = np.einsum('ij,ij', objects[:, :, r1], objects[:, :, r2].conj())
            ph_diff[r2, r1] = ph_diff[r1, r2].conj()

    if wtype == 'weighted':
        weights = np.abs(ph_diff)
    elif wtype == 'unweighted':
        weights = (np.abs(ph_diff) > threshold).astype(np.float)

    idx = weights > 0
    ph_diff[idx] = ph_diff[idx]/np.abs(ph_diff[idx])
    degree = np.sum(weights, axis=1)
    laplacian = np.diag(degree) - ph_diff * weights
    sig, v = eigsh(laplacian, 1, which='SM')
    idx = np.abs(v) > threshold
    phases = v
    phases[idx] = phases[idx]/np.abs(phases[idx])
    phases[~idx] = 1

    v *= v[0].conj()

    result = np.zeros((d1, d2), dtype=complex)
    count = np.zeros((d1, d2))
    for r in range(R):
        result[:, :] += phases[r].conj() * objects[:, :, r]
        count[:, :] += np.abs(objects[:, :, r]) > threshold

    result /= count
    return result



import math

import numpy as np
import numba
import scipy.constants as const


# Calculation of the relativistic electron wavelength in meters
def wavelength(U):
    '''
    Calculate the electron wavelength

    Parameters
    ----------

    U : float
        Acceleration voltage in kV

    Returns
    -------

    wavelength : float
        Wavelength in m

    Examples
    --------

    >>> wavelength(300)
    1.9687489006848795e-12
    '''
    e = const.elementary_charge  # Elementary charge  !!! 1.602176634×10−19
    h = const.Planck  # Planck constant    !!! 6.62607004 × 10-34
    c = const.speed_of_light  # Speed of light
    m_0 = const.electron_mass  # Electron rest mass

    T = e*U*1000
    lambda_e = h*c/(math.sqrt(T**2+2*T*m_0*(c**2)))
    return lambda_e


@numba.njit
def offset(s1, s2):
    o1, ss1 = s1
    o2, ss2 = s2
    return o2 - o1


@numba.njit
def shift_by(sl, shift):
    origin, shape = sl
    return (
        origin + shift,
        shape
    )


@numba.njit
def shift_to(s1, origin):
    o1, ss1 = s1
    return (
        origin,
        ss1
    )


@numba.njit
def intersection(s1, s2):
    o1, ss1 = s1
    o2, ss2 = s2
    # Adapted from libertem.common.slice
    new_origin = np.maximum(o1, o2)
    new_shape = np.minimum(
        (o1 + ss1) - new_origin,
        (o2 + ss2) - new_origin,
    )
    new_shape = np.maximum(0, new_shape)
    return (new_origin, new_shape)


@numba.njit
def get_shifted(arr_shape: tuple, tile_origin: tuple, tile_shape: tuple, shift: tuple):
    '''
    Calculate the slices to cut out a shifted part of a 2D source
    array and place it into a target array, including tiling support.

    This works with negative and positive integer shifts.
    '''
    # TODO this could be adapted for full sig, nav, n-D etc support
    # and included as a method in Slice?
    full_slice = (np.array((0, 0)), arr_shape)
    tileslice = (tile_origin, tile_shape)
    shifted = shift_by(tileslice, shift)
    isect = intersection(full_slice, shifted)
    if np.prod(isect[1]) == 0:
        return (
            np.array([(0, 0), (0, 0)]),
            np.array([0, 0])
        )
    # We measure by how much we have clipped the zero point
    # This is zero if we didn't shift into the negative region beyond the original array
    clip = offset(shifted, isect)
    # Now we move the intersection to (0, 0) plus the amount we clipped
    # so that the overlap region is moved by the correct amount, in total
    targetslice = shift_by(shift_to(isect, np.array((0, 0))), clip)
    start = targetslice[0]
    length = targetslice[1]
    target_tup = np.stack((start, start+length), axis=1)
    offsets = isect[0] - targetslice[0]
    return (target_tup, offsets)


def to_slices(target_tup, offsets):
    target_slice = tuple(slice(s[0], s[1]) for s in target_tup)
    source_slice = tuple(slice(s[0] + o, s[1] + o) for (s, o) in zip(target_tup, offsets))
    return (target_slice, source_slice)


def bounding_box(array):
    # Based on https://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array
    # But return values that work as start:stop slices
    rows = np.any(array, axis=1)
    cols = np.any(array, axis=0)
    if np.any(rows):
        y_min, y_max = np.where(rows)[0][[0, -1]]
        x_min, x_max = np.where(cols)[0][[0, -1]]
        return np.array(((y_min, y_max+1), (x_min, x_max+1)))
    else:
        return np.array([(0, 0), (0, 0)])



import functools

import numpy as np
import sparse
import numba

from libertem.masks import circular
from libertem.corrections.coordinates import identity

from ptychography40.reconstruction.common import get_shifted


@functools.lru_cache(None)
def empty_mask(mask_shape, dtype):
    '''
    Return an empty sparse mask

    Improve readability
    '''
    return sparse.zeros(mask_shape, dtype=dtype)


def mask_pair_subpix(cy, cx, sy, sx, filter_center, semiconv_pix):
    '''
    Calculate positive and negative trotter mask for circular illumination
    using a method with subpixel capability.

    Parameters
    ----------

    cy, cx : float
        Position of the optical axis on the detector in px, center of illumination
    sy, sx : float
        Trotter shift value in px
    filter_center : numpy.ndarray
        Center illumination, i.e. zero order disk. This has to be circular and match the radius
        semiconv_pix. It is just passed as an argument for efficientcy to avoid unnecessary
        recalculation.
    semiconv_pix : float
        Semiconvergence angle in measured in detector pixel, i.e. radius of the zero order disk.

    Returns
    -------

    mask_positive, mask_negative : numpy.ndarray
        Positive and negative trotter mask
    '''
    mask_shape = filter_center.shape

    filter_positive = circular(
        centerX=cx+sx, centerY=cy+sy,
        imageSizeX=mask_shape[1], imageSizeY=mask_shape[0],
        radius=semiconv_pix,
        antialiased=True
    )

    filter_negative = circular(
        centerX=cx-sx, centerY=cy-sy,
        imageSizeX=mask_shape[1], imageSizeY=mask_shape[0],
        radius=semiconv_pix,
        antialiased=True
    )
    mask_positive = filter_center * filter_positive * (filter_negative == 0)
    mask_negative = filter_center * filter_negative * (filter_positive == 0)
    return mask_positive, mask_negative


@numba.njit
def mask_tile_pair(center_tile, tile_origin, tile_shape, filter_center, sy, sx):
    '''
    Numerical work horse for :meth:`mask_pair_shift`, including tiling support.

    The tiling support could be used to calculate the mask stack on the fly,
    including support for UDF.process_tile().

    Parameters
    ----------

    center_tile : numpy.ndarray
        Tile cut out from :code:`filter_center` for re-use to increase efficiency
    tile_origin : tuple
        Origin of the tile to calculate
    tile_shape : tuple
        Shape of the tile
    filter_center : numpy.ndarray
        Center illumination, i.e. zero order disk.
    sy, sx : float
        Trotter shift value in px

    Returns
    -------
    mask_positive : numpy.ndarray
        Positive trotter tile
    target_tup_p : numpy.ndarray of int
        Start and stop indices per axis that were used for shifting the positive trotter tile.
    offsets_p : numpy.ndarray
        Offsets per axis that were used for shifting the positive trotter tile.
    mask_negative : numpy.ndarray
        Negative trotter tile
    target_tup_n : numpy.ndarray
        Start and stop indices per axis that were used for shifting the negative trotter tile.
    offsets_n : numpy.ndarray
        Offsets per axis that were used for shifting the negative trotter tile.
    '''
    sy, sx, = np.int(np.round(sy)), np.int(np.round(sx))
    positive_tile = np.zeros_like(center_tile)
    negative_tile = np.zeros_like(center_tile)
    # We get from negative coordinates,
    # that means it looks like shifted to positive
    target_tup_p, offsets_p = get_shifted(
        arr_shape=np.array(filter_center.shape),
        tile_origin=tile_origin,
        tile_shape=tile_shape,
        shift=np.array((-sy, -sx))
    )
    # We get from positive coordinates,
    # that means it looks like shifted to negative
    target_tup_n, offsets_n = get_shifted(
        arr_shape=np.array(filter_center.shape),
        tile_origin=tile_origin,
        tile_shape=tile_shape,
        shift=np.array((sy, sx))
    )

    sta_y, sto_y, sta_x, sto_x = target_tup_p.flatten()
    off_y, off_x = offsets_p
    positive_tile[sta_y:sto_y, sta_x:sto_x] = filter_center[
        sta_y+off_y:sto_y+off_y,
        sta_x+off_x:sto_x+off_x
    ]
    sta_y, sto_y, sta_x, sto_x = target_tup_n.flatten()
    off_y, off_x = offsets_n
    negative_tile[sta_y:sto_y, sta_x:sto_x] = filter_center[
        sta_y+off_y:sto_y+off_y,
        sta_x+off_x:sto_x+off_x
    ]

    mask_positive = center_tile * positive_tile * (negative_tile == 0)
    mask_negative = center_tile * negative_tile * (positive_tile == 0)

    return (mask_positive, target_tup_p, offsets_p, mask_negative, target_tup_n, offsets_n)


def mask_pair_shift(cy, cx, sy, sx, filter_center, semiconv_pix):
    '''
    Calculate positive and negative trotter mask using a fast shifting method.

    It has the same signature as :meth:`mask_pair_subpix` for easy changing
    between the methods. That means several parameters that are only relevant
    for the subpix method are ignored in this function.

    Parameters
    ----------

    cy, cx : float
        Ignored, given implicitly by filter_center
    sy, sx : float
        Trotter shift value in px
    filter_center : numpy.ndarray
        Center illumination, i.e. zero order disk.
    semiconv_pix : float
        Ignored, given implicitly by filter_center

    Returns
    -------

    mask_positive, mask_negative : numpy.ndarray
        Positive and negative trotter mask
    '''
    (mask_positive, target_tup_p, offsets_p,
    mask_negative, target_tup_n, offsets_n) = mask_tile_pair(
        center_tile=np.array(filter_center),
        tile_origin=np.array((0, 0)),
        tile_shape=np.array(filter_center.shape),

        filter_center=np.array(filter_center),
        sy=sy,
        sx=sx
    )
    return mask_positive, mask_negative


def generate_mask(cy, cx, sy, sx, filter_center, semiconv_pix,
                  cutoff, dtype, method='subpix'):
    '''
    Generate the trotter mask for a specific shift sy, sx

    Parameters
    ----------

    cy, cx : float
        Position of the optical axis on the detector in px, center of illumination
    sy, sx : float
        Trotter shift value in px
    filter_center : numpy.ndarray
        Center illumination, i.e. zero order disk. This has to be circular and match the radius
        semiconv_pix if :code:`method=subpix`. It is passed for re-use to avoid unnecessary
        recalculation
    semiconv_pix : float
        Semiconvergence angle in measured in detector pixel, i.e. radius of the zero order disk.
    cutoff : int
        Minimum number of pixels in the positive and negative trotter. This can be used to purge
        very small trotters to reduce noise.
    dtype : numpy dtype
        dtype to use for the mask
    method : str, optional
        Can be :code:`'subpix'`(default) or :code:`'shift'` to switch between
        :meth:`mask_pair_subpix` and :meth:`mask_pair_shift` to generate the trotter pair.

    Returns
    -------
    mask : sparse.COO
        Mask in sparse.pydata.org COO format

    '''
    mask_shape = filter_center.shape
    # 1st diffraction order and primary beam don't overlap
    if sx**2 + sy**2 > 4*np.sum(semiconv_pix**2):
        return empty_mask(mask_shape, dtype=dtype)

    if np.allclose((sy, sx), (0, 0)):
        # The zero order component (0, 0) is special, comes out zero with above code
        m_0 = filter_center / filter_center.sum()
        return sparse.COO(m_0.astype(dtype))

    params = dict(
        cy=cy, cx=cx, sy=sy, sx=sx,
        filter_center=filter_center,
        semiconv_pix=semiconv_pix,
    )

    if method == 'subpix':
        mask_positive, mask_negative = mask_pair_subpix(**params)
    elif method == 'shift':
        mask_positive, mask_negative = mask_pair_shift(**params)
    else:
        raise ValueError(f"Unsupported method {method}. Allowed are 'subpix' and 'shift'")

    non_zero_positive = mask_positive.sum()
    non_zero_negative = mask_negative.sum()

    if non_zero_positive >= cutoff and non_zero_negative >= cutoff:
        m = (
            mask_positive / non_zero_positive
            - mask_negative / non_zero_negative
        ) / 2
        m_dt = m.astype(dtype)
        m_sparse = sparse.COO(m_dt)
        return m_sparse
    else:
        # Exclude small, missing or unbalanced trotters
        return empty_mask(mask_shape, dtype=dtype)


def generate_masks(reconstruct_shape, mask_shape, dtype, lamb, dpix, semiconv,
        semiconv_pix, transformation=None, cy=None, cx=None, cutoff=1,
        cutoff_freq=np.float32('inf'), method='subpix'):
    '''
    Generate the trotter mask stack.

    The y dimension is trimmed to size(y)//2 + 1 to exploit the inherent
    symmetry of the mask stack.

    Parameters
    ----------

    reconstruct_shape : tuple(int)
        Shape of the reconstructed area
    mask_shape : tuple(int)
        Shape of the detector
    dtype : numpy dtype
        dtype to use for the mask stack
    lamb : float
        Wavelength of the illuminating radiation in m
    dpix : float or (float, float)
        Scan step in m. Tuple (y, x) in case scan step is different in x and y direction.
    semiconv : float
        Semiconvergence angle of the illumination in radians
    semiconv_pix : float
        Semiconvergence angle in measured in detector pixel, i.e. radius of the zero order disk.
    transformation : numpy.ndarray, optional
        Matrix for affine transformation from the scan coordinate directions
        to the detector coordinate directions. This does not include the scale, which is handled by
        dpix, lamb, semiconv and semiconv_pix. It should only be used to rotate and flip
        the coordinate system as necessary. See also
        https://github.com/LiberTEM/LiberTEM/blob/master/src/libertem/corrections/coordinates.py
    cy, cx : float, optional
        Position of the optical axis on the detector in px, center of illumination.
        Default: Center of the detector
    cutoff : int, optional
        Minimum number of pixels in the positive and negative trotter. This can be used to purge
        very small trotters to reduce noise. Default is 1, i.e. no cutoff unless one trotter is
        empty.
    cutoff_freq: float
        Trotters belonging to a spatial frequency higher than this value in reciprocal pixel
        coordinates will be cut off.
    method : str, optional
        Can be :code:`'subpix'`(default) or :code:`'shift'` to switch between
        :meth:`mask_pair_subpix` and :meth:`mask_pair_shift` to generate a trotter pair.

    Returns
    -------
    masks : sparse.COO
        Masks in sparse.pydata.org COO format. y and x frequency index are FFT shifted, i.e. the
        zero frequency is at (0,0) and negative frequencies are in the far quadrant and reversed.
        The y frequency index is cut in half with size(y)//2 + 1 to exploit the inherent symmetry
        of a real-valued Fourier transform. The y and x index are then flattened to make it
        suitable for using it with MaskContainer.
    '''
    reconstruct_shape = np.array(reconstruct_shape)

    dpix = np.array(dpix)

    d_Kf = np.sin(semiconv)/lamb/semiconv_pix
    d_Qp = 1/dpix/reconstruct_shape

    if cy is None:
        cy = mask_shape[0] / 2
    if cx is None:
        cx = mask_shape[1] / 2

    if transformation is None:
        transformation = identity()

    filter_center = circular(
        centerX=cx, centerY=cy,
        imageSizeX=mask_shape[1], imageSizeY=mask_shape[0],
        radius=semiconv_pix,
        antialiased=True
    )

    half_reconstruct = (reconstruct_shape[0]//2 + 1, reconstruct_shape[1])
    masks = []

    for row in range(half_reconstruct[0]):
        for column in range(half_reconstruct[1]):
            # Do an fftshift of q and p
            qp = np.array((row, column))
            flip = qp > (reconstruct_shape / 2)
            real_qp = qp.copy()
            real_qp[flip] = qp[flip] - reconstruct_shape[flip]

            if np.sum(real_qp**2) > cutoff_freq**2:
                masks.append(empty_mask(mask_shape, dtype=dtype))
                continue

            # Shift of diffraction order relative to zero order
            # without rotation in physical coordinates
            real_sy_phys, real_sx_phys = real_qp * d_Qp
            # We apply the transformation backwards to go
            # from physical orientation to detector orientation,
            # while the forward direction in center of mass analysis
            # goes from detector coordinates to physical coordinates
            # Afterwards, we transform from physical detector coordinates
            # to pixel coordinates
            sy, sx = ((real_sy_phys, real_sx_phys) @ transformation) / d_Kf

            masks.append(generate_mask(
                cy=cy, cx=cx, sy=sy, sx=sx,
                filter_center=filter_center,
                semiconv_pix=semiconv_pix,
                cutoff=cutoff,
                dtype=dtype,
                method=method,
            ))

    # Since we go through masks in order, this gives a mask stack with
    # flattened (q, p) dimension to work with dot product and mask container
    masks = sparse.stack(masks)
    return masks



import numpy as np
import numba

from libertem.udf import UDF
from libertem.common.container import MaskContainer

from ptychography40.reconstruction.ssb.trotters import generate_masks


@numba.njit(fastmath=True, cache=True, parallel=True, nogil=True)
def rmatmul_csc_fourier(n_threads, left_dense, right_data, right_indices, right_indptr,
                        coordinates, row_exp, col_exp, res_inout):
    '''
    Fold :meth:`SSB_UDF.merge_dot_result` into a sparse dot product from
    :meth:`libertem.common.numba._rmatmul_csc`

    The result can be directly merged into the result buffer instead of
    instantiating the intermediate dot result, which can be large. That way,
    this method can process entire memory-mapped partitions efficiently.
    Furthermore, it allows early skipping of trotters that are empty in this tile.

    It uses multithreading to process different parts of a tile in parallel.
    '''
    left_rows = left_dense.shape[0]
    p_size = col_exp.shape[1]
    q_size = row_exp.shape[1]
    # We subdivide in blocks per thread so that each thread
    # writes exclusively to its own part of an intermediate result buffer.
    # Using prange and automated '+=' merge leads to wrong results when using threading.
    blocksize = max(int(np.ceil(left_rows / n_threads)), 1)
    resbuf = np.zeros((n_threads, q_size, p_size), dtype=res_inout.dtype)
    # The blocks are processed in parallel
    for block in numba.prange(n_threads):
        start = block * blocksize
        stop = min((block + 1) * blocksize, left_rows)
        for left_row in range(start, stop):
            # Pixel coordinates in nav dimension
            y, x = coordinates[left_row]
            for q in range(q_size):
                for p in range(p_size):
                    # right_column is the mask index
                    right_column = q * p_size + p
                    # Descent into CSC data structure
                    offset = right_indptr[right_column]
                    items = right_indptr[right_column+1] - offset
                    if items > 0:
                        # We accumulate for the whole mask into acc
                        # before applying the phase factor
                        acc = 0
                        # Iterate over non-zero entries in this mask
                        for i in range(items):
                            index = i + offset
                            right_row = right_indices[index]
                            right_value = right_data[index]
                            acc += left_dense[left_row, right_row] * right_value
                        # Phase factor for this scan point and mask
                        factor = row_exp[y, q] * col_exp[x, p]
                        # Applying the factor, accumulate in per-thread buffer
                        resbuf[block, q, p] += acc * factor
    # Finally, accumulate per-thread buffer into result
    res_inout += np.sum(resbuf, axis=0)


class SSB_UDF(UDF):
    '''
    UDF to perform ptychography using the single side band (SSB) method :cite:`Pennycook2015`.
    '''
    def __init__(self, lamb, dpix, semiconv, semiconv_pix,
                 dtype=np.float32, cy=None, cx=None, transformation=None,
                 cutoff=1, method='subpix', mask_container=None,):
        '''
        Parameters
        ----------

        lamb: float
            The illumination wavelength in m. The function :meth:`ptychography40.common.wavelength`
            allows to calculate the electron wavelength as a function of acceleration voltage.
        dpix: float or Iterable(y, x)
            STEM pixel size in m
        semiconv: float
            STEM semiconvergence angle in radians
        dtype: np.dtype
            dtype to perform the calculation in
        semiconv_pix: float
            Diameter of the primary beam in the diffraction pattern in pixels
        cy, cx : float, optional
            Position of the optical axis on the detector in px, center of illumination.
            Default: Center of the detector
        transformation: numpy.ndarray() of shape (2, 2) or None
            Transformation matrix to apply to shift vectors. This allows to adjust for scan rotation
            and mismatch of detector coordinate system handedness, such as flipped y axis for MIB.
        cutoff : int
            Minimum number of pixels in a trotter
        method : 'subpix' or 'shift'
            Method to use for generating the mask stack
        mask_container: MaskContainer
            Allows to pass in a precomputed mask stack when using with single thread live data
            or with an inline executor as a work-around. The number of masks is sanity-checked
            to match the other parameters.
            The proper fix is https://github.com/LiberTEM/LiberTEM/issues/335
        '''
        super().__init__(lamb=lamb, dpix=dpix, semiconv=semiconv, semiconv_pix=semiconv_pix,
                         dtype=dtype, cy=cy, cx=cx, mask_container=mask_container,
                         transformation=transformation, cutoff=cutoff, method=method)

    EFFICIENT_THREADS = 4

    def get_result_buffers(self):
        ''
        dtype = np.result_type(np.complex64, self.params.dtype)
        component_dtype = np.result_type(np.float32, self.params.dtype)
        return {
            'fourier': self.buffer(
                kind="single", dtype=dtype, extra_shape=self.reconstruct_shape,
                where='device'
            ),
            'complex': self.buffer(
                kind="single", dtype=dtype, extra_shape=self.reconstruct_shape,
                use='result_only',
            ),
            'amplitude': self.buffer(
                kind="single", dtype=component_dtype, extra_shape=self.reconstruct_shape,
                use='result_only',
            ),
            'phase': self.buffer(
                kind="single", dtype=component_dtype, extra_shape=self.reconstruct_shape,
                use='result_only',
            ),
        }

    @property
    def reconstruct_shape(self):
        return tuple(self.meta.dataset_shape.nav)

    def get_task_data(self):
        ''
        # shorthand, cupy or numpy
        xp = self.xp

        if self.meta.device_class == 'cpu':
            backend = 'numpy'
        elif self.meta.device_class == 'cuda':
            backend = 'cupy'
        else:
            raise ValueError("Unknown device class")

        # Hack to pass a fixed external container
        # In particular useful for single-process live processing
        # or inline executor
        if self.params.mask_container is None:
            masks = generate_masks(
                reconstruct_shape=self.reconstruct_shape,
                mask_shape=tuple(self.meta.dataset_shape.sig),
                dtype=self.params.dtype,
                lamb=self.params.lamb,
                dpix=self.params.dpix,
                semiconv=self.params.semiconv,
                semiconv_pix=self.params.semiconv_pix,
                cy=self.params.cy,
                cx=self.params.cx,
                transformation=self.params.transformation,
                cutoff=self.params.cutoff,
                method=self.params.method,
            )
            container = MaskContainer(
                mask_factories=lambda: masks, dtype=masks.dtype,
                use_sparse='scipy.sparse.csr', count=masks.shape[0], backend=backend
            )
        else:
            container = self.params.mask_container
            target_size = (self.reconstruct_shape[0] // 2 + 1)*self.reconstruct_shape[1]
            container_shape = container.computed_masks.shape
            expected_shape = (target_size, ) + tuple(self.meta.dataset_shape.sig)
            if container_shape != expected_shape:
                raise ValueError(
                    f"External mask container doesn't have the expected shape. "
                    f"Got {container_shape}, expected {expected_shape}. "
                    "Mask count (self.meta.dataset_shape.nav[0] // 2 + 1) "
                    "* self.meta.dataset_shape.nav[1], "
                    "Mask shape self.meta.dataset_shape.sig. "
                    "The methods generate_masks_*() help to generate a suitable mask stack."
                )

        # Precalculated LUT for Fourier transform
        # The y axis is trimmed in half since the full trotter stack is symmetric,
        # i.e. the missing half can be reconstructed from the other results
        row_steps = -2j*np.pi*np.linspace(0, 1, self.reconstruct_shape[0], endpoint=False)
        col_steps = -2j*np.pi*np.linspace(0, 1, self.reconstruct_shape[1], endpoint=False)

        half_y = self.reconstruct_shape[0] // 2 + 1
        full_x = self.reconstruct_shape[1]

        # This creates a 2D array of row x spatial frequency
        row_exp = np.exp(
            row_steps[:, np.newaxis]
            * np.arange(half_y)[np.newaxis, :]
        )
        # This creates a 2D array of col x spatial frequency
        col_exp = np.exp(
            col_steps[:, np.newaxis]
            * np.arange(full_x)[np.newaxis, :]
        )

        steps_dtype = np.result_type(np.complex64, self.params.dtype)

        return {
            "masks": container,
            "row_exp": xp.array(row_exp.astype(steps_dtype)),
            "col_exp": xp.array(col_exp.astype(steps_dtype)),
            "backend": backend
        }

    def merge(self, dest, src):
        ''
        dest.fourier[:] += src.fourier

    def merge_dot_result(self, dot_result):
        # shorthand, cupy or numpy
        xp = self.xp

        tile_depth = dot_result.shape[0]

        # We calculate only half of the Fourier transform due to the
        # inherent symmetry of the mask stack. In this case we
        # cut the y axis in half. The "+ 1" accounts for odd sizes
        # The mask stack is already trimmed in y direction to only contain
        # one of the trotter pairs
        half_y = self.results.fourier.shape[0] // 2 + 1

        # Get the real x and y indices within the dataset navigation dimension
        # for the current tile
        y_indices = self.meta.coordinates[:, 0]
        x_indices = self.meta.coordinates[:, 1]

        # This loads the correct entries for the current tile from the pre-calculated
        # 1-D DFT matrices using the x and y indices of the frames in the current tile
        # fourier_factors_row is already trimmed for half_y, but the explicit index
        # is kept for clarity
        fourier_factors_row = self.task_data.row_exp[y_indices, :half_y, np.newaxis]
        fourier_factors_col = self.task_data.col_exp[x_indices, np.newaxis, :]

        # The masks are in order [row, col], but flattened. Here we undo the flattening
        dot_result = dot_result.reshape((tile_depth, half_y, self.results.fourier.shape[1]))

        # Calculate the part of the Fourier transform for this tile.
        # Reconstructed shape corresponds to depth of mask stack, see above.
        # Shape of operands:
        # dot_result: (tile_depth, reconstructed_shape_y // 2 + 1, reconstructed_shape_x)
        # fourier_factors_row: (tile_depth, reconstructed_shape_y // 2 + 1, 1)
        # fourier_factors_col: (tile_depth, 1, reconstructed_shape_x)
        # The einsum is equivalent to
        # (dot_result*fourier_factors_row*fourier_factors_col).sum(axis=0)
        # The product of the Fourier factors for row and column implicitly builds part
        # of the full 2D DFT matrix through NumPy broadcasting
        # The sum of axis 0 (tile depth) already performs the accumulation for the tile
        # stack before patching the missing half for the full result.
        # Einsum is about 3x faster in this scenario, likely because of not building a large
        # intermediate array before summation
        self.results.fourier[:half_y] += xp.einsum(
            'i...,i...,i...',
            dot_result,
            fourier_factors_row,
            fourier_factors_col
        )

    def postprocess(self):
        ''
        # shorthand, cupy or numpy
        xp = self.xp
        half_y = self.results.fourier.shape[0] // 2 + 1
        # patch accounts for even and odd sizes
        # FIXME make sure this is correct using an example that transmits also
        # the high spatial frequencies
        patch = self.results.fourier.shape[0] % 2
        # We skip the first row since it would be outside the FOV
        extracted = self.results.fourier[1:self.results.fourier.shape[0] // 2 + patch]
        # The coordinates of the bottom half are inverted and
        # the zero column is rolled around to the front
        # The real part is inverted
        self.results.fourier[half_y:] = -xp.conj(
            xp.roll(xp.flip(xp.flip(extracted, axis=0), axis=1), shift=1, axis=1)
        )

    def get_results(self):
        '''
        Since we derive the wave front with a linear function from intensities,
        but the wave front is inherently an amplitude,
        we take the square root of the calculated amplitude and
        combine it with the calculated phase.
        '''
        inverse = np.fft.ifft2(self.results.fourier)
        amp = np.sqrt(np.abs(inverse))
        phase = np.angle(inverse)
        return {
            'fourier': self.results.fourier,
            'complex': amp * np.exp(1j*phase),
            'amplitude': amp,
            'phase': phase,
        }

    def process_tile(self, tile):
        ''
        # We flatten the signal dimension of the tile in preparation of the dot product
        tile_flat = tile.reshape(tile.shape[0], -1)
        if self.task_data.backend == 'cupy':
            # Load the preconditioned trotter stack from the mask container:
            # * Coutout for current tile
            # * Flattened signal dimension
            # * Clean CSR matrix
            # * On correct device
            masks = self.task_data.masks.get(
                self.meta.slice, transpose=False, backend=self.task_data.backend,
                sparse_backend='scipy.sparse.csr'
            )
            # Skip an empty tile since the contribution is 0
            if masks.nnz == 0:
                return
            # This performs the trotter integration
            # As of now, cupy doesn't seem to support __rmatmul__ with sparse matrices
            dot_result = masks.dot(tile_flat.T).T
            self.merge_dot_result(dot_result)
        else:
            # We calculate only half of the Fourier transform due to the
            # inherent symmetry of the mask stack. In this case we
            # cut the y axis in half. The "+ 1" accounts for odd sizes
            # The mask stack is already trimmed in y direction to only contain
            # one of the trotter pairs
            half_y = self.results.fourier.shape[0] // 2 + 1
            tpw = self.meta.threads_per_worker
            if (tpw is None) or (tpw >= self.EFFICIENT_THREADS):
                # Load the preconditioned trotter stack from the mask container:
                # * Coutout for current tile
                # * Flattened signal dimension
                # * Transposed for right hand side of dot product
                #   (rmatmul_csc_fourier() takes care of putting things where they belong
                #    in the result)
                # * Clean CSC matrix
                # * On correct device (CPU)
                masks = self.task_data.masks.get(
                    self.meta.slice, transpose=True, backend=self.task_data.backend,
                    sparse_backend='scipy.sparse.csc'
                )
                # Skip an empty tile since the contribution is 0
                if masks.nnz == 0:
                    return

                # This combines the trotter integration with merge_dot_result
                # into a Numba loop that eliminates a potentially large intermediate result
                # and allows efficient multithreading on a large tile
                rmatmul_csc_fourier(
                    n_threads=self.meta.threads_per_worker,
                    left_dense=tile_flat,
                    right_data=masks.data,
                    right_indices=masks.indices,
                    right_indptr=masks.indptr,
                    coordinates=self.meta.coordinates,
                    row_exp=self.task_data.row_exp,
                    col_exp=self.task_data.col_exp,
                    res_inout=self.results.fourier[:half_y]
                )
            else:
                masks = self.task_data.masks.get(
                    self.meta.slice, transpose=False, backend=self.task_data.backend,
                    sparse_backend='scipy.sparse.csr'
                )
                # Skip an empty tile since the contribution is 0
                if masks.nnz == 0:
                    return
                # This performs the trotter integration
                # Transposed copy of the input data for fast scipy.sparse __matmul__()
                dot_result = masks.dot(tile_flat.T.copy()).T
                self.merge_dot_result(dot_result)

    def get_backends(self):
        ''
        return ('numpy', 'cupy')

    def get_tiling_preferences(self):
        ''
        dtype = np.result_type(np.complex64, self.params.dtype)
        result_size = np.prod(self.reconstruct_shape) * dtype.itemsize
        if self.meta.device_class == 'cuda':
            free, total = self.xp.cuda.runtime.memGetInfo()
            total_size = min(100e6, free // 4)
            good_depth = max(1, total_size / result_size * 4)
            return {
                "depth": good_depth,
                "total_size": total_size,
            }
        else:
            tpw = self.meta.threads_per_worker
            if (tpw is None) or (tpw >= self.EFFICIENT_THREADS):
                # The parallel Numba loop can process entire partitions efficiently
                # since it accumulates in the result without large intermediate
                # data
                return {
                    "depth": self.TILE_DEPTH_MAX,
                    "total_size": self.TILE_SIZE_MAX,
                }
            else:
                # We limit the depth of a tile so that the intermediate
                # results from processing a tile fit into the CPU cache.
                good_depth = max(1, 1e6 / result_size)
                return {
                    "depth": int(good_depth),
                    # We reduce the size of a tile since a transposed copy of
                    # each tile will be taken to speed up the sparse matrix product
                    "total_size": 0.5e6,
                }



import re
import os
import codecs
import subprocess
from setuptools.command.sdist import sdist
from setuptools.command.build_py import build_py
from setuptools import setup


class BakedRevisionBuilderSdist(sdist):
    def make_release_tree(self, base_dir, files):
        if not self.dry_run:
            write_baked_revision(base_dir)
        sdist.make_release_tree(self, base_dir, files)


class BakedRevisionBuilderBuildPy(build_py):
    def run(self):
        if not self.dry_run:
            write_baked_revision(self.build_lib)
        build_py.run(self)


def mkpath(path):
    if not os.path.exists(path):
        os.makedirs(path)


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r').read()


def remove_rst_roles(txt):
    return re.sub(':(cite|doc):`[^`]+` ?', '', txt)


def get_git_rev():
    # NOTE: this is a copy from src/ptychography40/versioning.py
    # this is because it is not guaranteed that we can import our own packages
    # from setup.py AFAIK
    try:
        new_cwd = os.path.abspath(os.path.dirname(__file__))
        rev_raw = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=new_cwd)
        return rev_raw.decode("utf8").strip()
    except Exception:
        return "unknown"


def write_baked_revision(base_dir):
    dest_dir = os.path.join(base_dir, 'ptychography40')
    baked_dest = os.path.join(dest_dir, '_baked_revision.py')
    mkpath(dest_dir)

    with open(baked_dest, "w") as f:
        f.write(r'revision = "%s"' % get_git_rev())


def find_version(*file_paths):
    """
    "stolen" from pip's setup.py
    """
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="ptychography40",
    version=find_version("src", "ptychography40", "__version__.py"),
    license='GPL v3',
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.6',
    install_requires=[
        "numpy",
        "scipy",
        "sparse",
        "libertem>=0.7.0",
        # Pinned due to https://github.com/pydata/sparse/issues/257
        # Ensure compatibility with numpy 1.17
        "numba>=0.46",
    ],
    extras_require={
    },
    package_dir={"": "src"},
    packages=[
        "ptychography40",
    ],

    cmdclass={
        'sdist': BakedRevisionBuilderSdist,
        'build_py': BakedRevisionBuilderBuildPy,
    },
    keywords="Ptychography",
    description="Phase reconstruction using ptychography",
    long_description=remove_rst_roles(read("README.rst")),
    long_description_content_type="text/x-rst",
    url="https://ptychography-4-0.github.io/ptychography/",
    author_email="ptychography4@listen.helmholtz-muenchen.de",
    author="the Ptychography 4.0 team",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)
